#include <atomic>
#include <cassert>
#include <cstddef>
#include "fib.h"

#ifndef FIB_MAX_INSTANCES
#define FIB_MAX_INSTANCES 2
#endif /* ifndef MAX_FIB_INSTANCES */

namespace Fib {
class CorePrivate {
 public:
CorePrivate() {
    mutex_.clear(std::memory_order_release);
    iteration_ = 0;
    t1_ = 0;
    t2_ = 1;
}

~CorePrivate() {}

// Each successive call will return the next number
// in the Fibonacci sequence (0,1,1,2,3,5,8,...)
int CalcNext() {
    int next;

    Lock();

    switch (iteration_++) {
        case 0:
            next = t1_;
            break;
        case 1:
            next = t2_;
            break;
        default:
            next = t1_ + t2_;
            t1_ = t2_;
            t2_ = next;
    }

    Unlock();

    return next;
}

// Restart Fibonacci sequence
void Restart() {
    mutex_.clear(std::memory_order_release);
    iteration_ = 0;
    t1_ = 0;
    t2_ = 1;
}

bool GetInUse() {
    return in_use_;
}

void SetInUse(bool val) {
    in_use_ = val;
}

 private:
void Lock() {
    while (mutex_.test_and_set(
        std::memory_order_acquire)) {}
}

void Unlock() {
    mutex_.clear(std::memory_order_release);
}

bool in_use_;
std::atomic_flag mutex_;
int iteration_;
int t1_;
int t2_;
};

static CorePrivate instances[FIB_MAX_INSTANCES];

Core::Core() {
    d_ptr_ = NULL;

    for (int i = 0; i < FIB_MAX_INSTANCES; i++) {
        if (!instances[i].GetInUse()) {
            instances[i].SetInUse(true);
            d_ptr_ = &(instances[i]);
            return;
        }
    }

    bool available_instances = false;
    assert(available_instances = false);
}
Core::~Core() = default;

// Reset Fibonacci sequence
void Core::Restart() {
    d_ptr_->Restart();
}

// Each successive call will return the next number
// in the Fibonacci sequence (0,1,1,2,3,5,8,...)
int Core::CalcNext() {
    return d_ptr_->CalcNext();
}
}  // namespace Fib
